//////////////////////////////////////////////////////////////////////////////
//
// Module:
//  mem.h
//
// Description:
//  This module implements prototype, defines and structures needed
//  by the mem.c module of the MMULib.
//
// Author:
//  (c) 2004 Jouni 'Mr.Spiv' Korhonen
//
// Version:
//  v0.1 - 9-May-2004 Initial release version.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef _mem_h_included
#define _mem_h_included

#ifdef __cplusplus
extern "C" {
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  struct customMMU
//
// Members:
//  heapStart - In a case the program wants to setup a custom MMU table, it
//              is possible to define the start of the heap for libc's
//              memory management system. This is especially useful in cases
//              when there is more than 8M of RAM available in the GP32.
//              Set to NO_HEAP if not used.
//  heapEnd   - In a case the program wants to setup a custom MMU table, it
//              is possible to define the end of the heao for libc's
//              memory management system.
//              Set to NULL if not used.
//  ttb       - A ptr to memory where the custom MMU table should be placed.
//              The ptr must point to a physical memory ( a flat mapped
//              region) as the MMU is turned off when new MMU tables get
//              programmed. This ptr has no effect if the mmu tables are not
//              defined (mmu prt is NULL).
//  ttbsize   - Amount of the memory in octets reserved for the new MMU table.
//  mmu       - A ptr to the new MMU table description. See the module 
//              mmu.h how MMU tables are defined.
//  domain    - Set Domain control register to the given value. If the value
//              is NODOMAIN the domain control register is not changed.
//  pid       - Set PID control register to the given value. If the value is
//              NOPID the PID control register is not changed.
//  S         - Set the S control register bit to the given value. If the
//              value is NO_S the S control register bit is not changed.
//  R         - Set the R control register bit to the given value. If the
//              value is NO_R the R control register bit is not changed.
//
// Note:
//  heapStart & heapEnd are dangerous to set to non-NO_HEAP values if there is
//  only thye standard 8M RAM available. In that case it is better to rely on
//  libc's default way of managing memory allocations.
//
//////////////////////////////////////////////////////////////////////////////

struct customMMU {
	char *heapStart;
	char *heapEnd;
	char *ttb;
	int ttbSize;
	unsigned long *mmu;
	unsigned long domain;
	unsigned long pid;
	int S,R;
};

//
// Defines used along with the struct customMMU
//

#define NODOMAIN	0xaaaaaaaa
#define NOPID		0xffffffff
#define NO_R		-1
#define NO_S		-1
#define NO_HEAP		0x00000000

// Error return codes 

#define ERR_MEM_LEVEL1	-666		// non compatible Level1 descriptors
#define ERR_MEM_LEVEL2	-668		// non compatible level2 descriptors
#define ERR_MEM_EXCEPTIONS	-669	// exception handlers have been
                                        // patched already
//
// These defines are used to define the Exception Vector the program
// want to change with the installExceptionHandler() function.
//

#define EXC_RESET	0
#define EXC_UNDEF	1
#define EXC_SWI		2
#define EXC_PABORT	3
#define EXC_DABORT	4
#define EXC_IRQ		6
#define EXC_FIQ		7
#define EXC_MAGIC_OFFSET	180    // Internal stuff

//
//
//
//
//
//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  This function returns the amount of available physical RAM in megabytes.
//
// Parameters:
//  None.
//
// Returns:
//  The available physical RAM in megabytes.
//
// Note:
//  The function disables/enables the MMU and flushed both instruction and
//  data caches. This function also modifies memory above 0x0c800000 so you
//  are advised to do it at the very beginning of you code when no memory
//  allocation what so ever has not yet taken place!!!
//  You should call this function when interrupts are disabled.
//
//////////////////////////////////////////////////////////////////////////////

int availPhysMem( void );


//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  This function initializes the MMULib and sets up required patches and
//  MMU tables. This function should be called _only_ once and some of the
//  installed patches are "permanent" for the lifetime of the program.
//
// Parameters:
//  allowSDK - A boolean and if set to 1 the patched libc memory management
//             becomes aware of the Gamepark SDK. It is possible to even
//             mix both libc and Gamepark SDK memory management functions.
//             Refer the table below for possible combinations.
//             Set to -1 if you want just ignore this parameter e.g. in a 
//             case when you want to leave this for crt0.o.
//
//  allowIRQ - A boolean and if set to 1 it is possible to replace BIOS
//             defined Exception Handlers. Either directly writing to ROM
//             area (strongly discouraged) or using the library provided
//             installExceptionHandler() function (strongly preferred).
//             Note that this option _should_ not be used if the caller
//             defines custom MMU tables.
//
//  mmu      - A ptr to customMMU structure defined custom MMU tables
//             and other low level memory management details. If a NULL
//             pointer is passed then BIOS set up MMU tables are used.
//
// Returns:
//  0 if ok, or ERR_MEM_LEVEL1 if patching Expection Vectors failed at 
//  Level1 or ERR_MEM_LEVEL2 if the error happened with Level2 
//  descriptors. ERR_MEM_EXCEPTIONS if Exception Vectors were patched
//  already.
//
// Table for SDK & libc memory management combinations:
//
//  allowSDK  available-RAM  outcome
//    0         8M             Only libc malloc() and stuff works reliably.
//    0         32/64M         Libc malloc() and stuff return memory from
//                             the above 8M area. Low 8M is not used by the
//                             libc.
//    1         8M             Only Gamepark SDK gm_malloc() and stuff works.
//                             Libc malloc() etc always fail.
//    1         32/64M         Gamepark SDK gm_malloc() and stuff return
//                             memory from the low 8M area. Libc return
//                             memory above 8M area.
//
// Note:
//  This function disables/enables the MMU. The function _should_ be called
//  when the BIOS set up MMU table is still valid and functioning, unless
//  the caller has very good knowlegde what is going on.
//  This function should be called from a RAM area that has originally
//  flat mapping (if you don't know what that means this library is not
//  for you).
//
//  This function (and the MMULib in general) expects support for enhanced
//  memory management. Namely the architecture specific file syscalls.o
//  must have been patched (sbrk funtion) to understand the difference 
//  between the libc and Gamepark SDK. If the libc has not been patched
//  then the "allowSDK" option is useless and libc & Gamepark SDK memory
//  management functions cannot be mixed.
//
//////////////////////////////////////////////////////////////////////////////

int initMMULib( int allowSDK, int allowIRQ, struct customMMU *mmu );


//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  This function is used to install custom Exception Handlers. In order to 
//  this function to work the "allowIRQ" paramater must be set true in the
//  initMMULib() function.
//
// Parameters:
//  vector - Defines the Exception Vector to patch.
//  func   - A ptr to the Exception Handler function. Note that this handler
//           has to implement all required IRQ dispatching functionality and
//           handler the proper handshake with Interrupt Controller register.
//
// Returns:
//  0 if ok, -1 if error.
//
//////////////////////////////////////////////////////////////////////////////

int installExceptionHandler( int vector, void (*func)( void ) );


//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  This function is used to modify caching of _mapped_ memory areas.
//  All possible Page sizes and Sections are supported.
//
// Parameters:
//  start - the starting virtual address
//  end   - the ending virtual address
//  flags - caching related flags as defined in mmu.h
//
// Returns:
//  0 if ok, -1 if virtual addresses were invalid and -2 if the memory range
//  define by start and end contained non-mapped regions.
//
// Note: Invalidates TLBs, writeback and data caches.
//
//////////////////////////////////////////////////////////////////////////////

int modifyMMUCaching( void *start, void *end, unsigned long flags );


//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  This function returns either Level1 or Level2 MMU table descriptor for  
//  a given virtual address. If the memory pointed by the virtual address
//  has not been previously mapped an error will be returned.
//
// parameters:
//  vaddr - [in] virtual address.
//  level - [out] a ptr to interger. The level of returned MMU table
//          descriptor is returned here.
//
// Returns:
//  Descriptor if ok, FAULT_LVL1 if the vaddr was invalid. Note that error
//  is _also_ returned for non-mapped memory.
//  The integer pointed by "level" contains 1 if the returned descriptor is
//  a Level1 descriptor and the "level" contais 2 if the the returned
//  descriptor is a Level2 descriptor. In case of error the contents of
//  "level" is undefined.
//
//////////////////////////////////////////////////////////////////////////////

unsigned long getDescriptor( void *vaddr, int *level );


//////////////////////////////////////////////////////////////////////////////
//
// Description:
//  this function sets/modifies a Level1 or Level2 MMU table
//  descriptor to a given value.
//
// parameters:
//  vaddr - Virtual address whose MMU Level1 or Level2 descptor is to
//          be changed. It is up to the caller to "know" whether the
//          descriptor is Level1 or Level2.
//  desc  - Descriptor to be set.
//
// Returns:
//  0 if write was ok, -1 if the virtual addr was invalid.
//  
// Note:
//  TTB entries will be flushed as well as data cache and writeback.
//  Better call this when interrupts are disabled.
//
//////////////////////////////////////////////////////////////////////////////

int setDescriptor( void *vaddr, unsigned long desc );

//////////////////////////////////////////////////////////////////////////////
//
// The defines listed below describe the possible MMU table tags a program
// can use to construct custom MMU tables.
//
//
// [GLOBALS]
//
//   GBL_DOMAIN, <domain>
//   GBL_AP, AP_RORO|AP_RW_NA|AP_RWRO|AP_RWRW
//   GBL_CB, NCNB|CNB|NCB|CB
//
// [LEVEL1]
//
//   LVL1,((COARSE|FINE|FAULT,<start>,<end>)|
//         (SECTION,<start>,<end>,<phys>)),[LVL1_FLAGS],
//        LVL2,<start>,<end>,<phys>,LARGE|SMALL|TINY|FAULT,[LVL2_FLAGS]
//        LVL2,
//        LVL2,
//        ...
//   LVL1,...
//   MMU_END
//
// [LVL1_FLAGS]
//   LVL1_DOMAIN,<#>
//   LVL1_AP, AP_RORO|AP_RWNA|AP_RWRO|AP_RWRW
//   LVL1_CB, NCNB|CNB|NCB|CB
//
// [LVL2_FLAGS]
//   LVL2_CB, NCNB|CNB|NCB|CB
//   LVL2_AP0, AP_RORO|AP_RWNA|AP_RWRO|AP_RWRW
//   LVL2_AP1, AP_RORO|AP_RWNA|AP_RWRO|AP_RWRW
//   LVL2_AP2, AP_RORO|AP_RWNA|AP_RWRO|AP_RWRW
//   LVL2_AP3, AP_RORO|AP_RWNA|AP_RWRO|AP_RWRW
//
// Note that with SECTION Level1 descriptors the program does not need to
// define any Level2 descriptors.
//
// See the Example below, which does a valid MMU table for GP32 with
// 32M RAM and BIOS/FW remapped to 0x0c800000.
//
//
#if 0

static unsigned long _gp32_std_mmu[] = {
  GBL_DOMAIN, 0xf,   // Define "global" domain
  LVL1,COARSE,0x0,0xfffff,
    LVL2,0x0,0x7ffff,0xc800000,LARGE,LVL2F_AP,AP_RORO,LVL2F_CB,CNB,
  LVL1,COARSE,0xc000000,0xc800000,
    LVL2,0xc000000,0xc6fffff,0xc000000,SMALL,LVL2F_AP,AP_RWRW,LVL2F_CB,CB,
    LVL2,0xc700000,0xc7fffff,0xc700000,SMALL,LVL2F_AP,AP_RWRW,LVL2F_CB,NCNB,
  LVL1,COARSE,0xc800000,0xdffffff,
    LVL2,0xc800000,0xdffffff,0xc800000,SMALL,LVL2F_AP,AP_RWRW,LVL2F_CB,CB,
  LVL1,SECTION,0x14000000,0xe9ffffff,0x14000000,LVL1F_AP,AP_RWRW,LVL1F_CB,NCNB,
  MMU_DONE	
};
#endif   // Example


#ifdef __cplusplus
}
#endif
#endif
















